from __future__ import absolute_import, print_function
import importlib
import inspect
from gaw import GawServer
from gaw.entrypoint import Entrypoint
from argparse import ArgumentParser
import itertools
import sys
import signal
import os

def on_sigint(signal, frame):
    print('Exiting...')
    sys.exit(0)

def servicerunner_cli():
    signal.signal(signal.SIGINT, on_sigint)

    # add the executor path to the pythonpath
    sys.path.insert(0, os.getcwd())

    parser = ArgumentParser(description='Gaw Service Runner - CLI')
    parser.add_argument('module_name', type=str)
    parser.add_argument('--service', type=str)
    parser.add_argument('--port', type=int, default=5555)
    parser.add_argument('--ip', type=str, default='127.0.0.1')
    parser.add_argument('--secret', type=str)
    parser.add_argument('--is_encrypt', action='store_true')
    parser.add_argument('--kwargs', type=str)
    args = parser.parse_args()

    print('Gaw Service Runner - CLI')

    # load the designated module
    module_name = args.module_name
    module = importlib.import_module(module_name)

    # get a list of all gaw services in the module
    def is_gaw_service(cls):
        entry_points = Entrypoint.get_entrypoints_from_class(cls)
        return len(entry_points) > 0

    gaw_services = []
    for name, obj in inspect.getmembers(module):
        if not inspect.isclass(obj):
            continue

        if is_gaw_service(obj):
            service_name = getattr(obj, 'name')
            gaw_services.append((service_name, obj))

    # filter down according to the --service argument
    def filter_service(service):
        name, cls = service
        if args.service:
            return name == args.service
        else:
            return True

    selected_services = list(filter(filter_service, gaw_services))

    if len(selected_services) == 0:
        print('gawrunner: no service to be run, exiting...')
        return

    # run the service (try many ports in sequential fashion)
    for port_shifting in itertools.count():
        port = args.port + port_shifting
        try:
            # start the server
            server = GawServer(args.ip, port, secret=args.secret, is_encrypt=args.is_encrypt)
            for name, service_class in selected_services:
                if args.kwargs:
                    kwargs = eval('dict({})'.format(args.kwargs))
                    print('gawrunner: run {} service with parameters {}'.format(name, kwargs))
                    server.add(service_class, **kwargs)
                else:
                    server.add(service_class)
            server.run()
            # this is important!
            # forget to 'break' here, you cannot stop the programm from keeping running
            break
        except OSError as err:
            if err.errno == 48:
                print('gawrunner: [err 48] port {} taken ... trynig another'.format(port))
                continue
            else:
                raise err

if __name__ == '__main__':
    servicerunner_cli()


from __future__ import print_function, absolute_import
from gaw import GawServer, entrypoint

# class TestService(object):
#
#     name = 'test_service'
#
#     def __init__(self):
#         print('init a service')
#         pass
#
#     @entrypoint
#     def plus(self, a, b):
#         return a + b
#
#     @entrypoint
#     def multiply(self, a, b):
#         return a * b
#
# service = Microservice('127.0.0.1', 4444, verbose=False)
# service.add(TestService).run()

class MathService(object):
    name = 'math_service'

    def __init__(self, hello_message):
        self.hello = hello_message

    @entrypoint
    def plus(self, a, b):
        return '{}: {}'.format(self.hello, a + b)

    @entrypoint
    def multiply(self, a, b):
        return '{}: {}'.format(self.hello, a * b)


service = GawServer('127.0.0.1', 5555, verbose=True, secret='Qx9XFxN17+zkUdcBIGZ0A1sQTkUSP4SZ', is_encrypt=True)
service.add(MathService, hello_message='Hello!')
service.run()




from __future__ import absolute_import
from gaw.postoffice.server import PostofficeServer

def handle(message):
    return 'success: {}'.format(message)

PostofficeServer(ip='0.0.0.0',
                 port=4444,
                 on_message=handle,
                 secret='anfJdufzm2FSbQca0dD8RgN01SO4Pgn0',
                 is_encrypt=True,
                 verbose=True)


from __future__ import absolute_import
from gaw.jsonsocketserver.client import JsonSocketClient
from gaw.jsonsocketserver.exceptions import JsonSocketException
import time

ip = '127.0.0.1'
port = 4444

client = JsonSocketClient(verbose=True)

# test client side error handling
try:
    print(client.request(ip, port, 'home', dict(), secret='Qx9XFxN17+zkUdcBIGZ0A1sQTkUSP4SZ', is_encrypt=True))
except JsonSocketException as e:
    print(e.name)
    print(e.trace)

print(client.request(ip, port, 'plus', dict(a=10, b=20), secret='Qx9XFxN17+zkUdcBIGZ0A1sQTkUSP4SZ', is_encrypt=True))

# times = 10000
# start_time = time.time()
# for i in range(times):
#     client.request(ip, port, 'home', '')
# end_time = time.time()
#
#
# print('taken:', (end_time - start_time) * 1000 / times)


from __future__ import print_function
from gaw.jsonsocketserver.server import JsonSocketServer

def home(path):
    print('home!')
    print('path:', path)
    raise ValueError('test error')

def plus(a, b, path):
    print('plus!')
    print('path:', path)
    return a + b

server = JsonSocketServer('0.0.0.0', 4444, secret='Qx9XFxN17+zkUdcBIGZ0A1sQTkUSP4SZ', is_encrypt=True)
server.register_route('home', home)
server.register_route('plus', plus)
server.start()


from __future__ import absolute_import
from gaw.postoffice.client import PostofficeClient

client = PostofficeClient(ip='127.0.0.1',
                          port=4444,
                          secret='anfJdufzm2FSbQca0dD8RgN01SO4Pgn0',
                          is_encrypt=True,
                          verbose=True)

response = client.send(dict(a='aoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuaoeuoa'))
print('response:', response)

import time

rounds = 10000
start_time = time.time()
for i in range(rounds):
    response = client.send(dict(a=u'aoeu'))
end_time = time.time()

print('each:', (end_time - start_time) / float(rounds) * 1000, 'ms')


from __future__ import print_function, absolute_import
from gaw.jsonsocketserver.datatype import RequestDataType, ResponseDataType
from gaw.postoffice import PostofficeClient
from threading import Thread
from gaw.jsonsocketserver.exceptions import JsonSocketException
from gaw.postoffice.exceptions import PostofficeException
import uuid
import datetime
import time
import eventlet
import traceback

class JsonSocketClient:

    def __init__(self, client_lifetime=30, verbose=False):
        self.client_lifetime = client_lifetime
        self.verbose = verbose

        self._client_pool = dict()
        self._is_busy = dict()
        self._last_act_at = dict()

        self._cleaner = Thread(target=self._cleaner)
        self._cleaner.daemon = True
        self._cleaner.start()

    def request(self, ip, port, path, payload, secret=None, is_encrypt=False):
        # may retry many times
        while True:
            client = self._get_client(ip, port, secret, is_encrypt)
            assert isinstance(client, PostofficeClient)

            if self.verbose:
                print('jsonsocketclient: requesting ip', ip, 'port', port, 'path', path, 'payload', payload)

            request = RequestDataType(id=uuid.uuid1().int,
                                      path=path,
                                      payload=payload)

            try:
                raw_response = client.send(request.dict())
            except PostofficeException as e:
                raise e
            except Exception as err:
                print('jsonsocketclient: connection error retrying ...')
                print('jsonsocketclient: err:', err, 'trace:', traceback.format_exc())
                # delete the old client
                self._delete(self._key(ip, port, secret, is_encrypt))
                eventlet.sleep(1)
                # retry
                continue

            if self.verbose:
                print('jsonsocketclient: response ', raw_response)

            response = ResponseDataType.parse(raw_response)

            if response.resp_to != request.id:
                raise ValueError('request error: wrong response id')

            if not response.success:
                exception = response.payload
                name = exception['name']
                message = exception['message']
                trace = exception['trace']
                raise JsonSocketException(name=name, message=message, trace=trace)

            return response.payload

    # private
    def _key(self, ip, port, secret, is_encrypt):
        return '{}:{}:{}:{}'.format(ip, port, secret, is_encrypt)

    def _init_client(self, ip, port, secret, is_encrypt):
        # may retry connection many times
        while True:
            try:
                if self.verbose:
                    print('jsonsocketclient: init client ip: {} port: {} secret: {} is_encrypt: {}'.format(ip, port, secret, is_encrypt))
                return PostofficeClient(ip, port, verbose=self.verbose,
                                        secret=secret,
                                        is_encrypt=is_encrypt)
            except Exception:
                print('jsonsocketclient: host is down retrying ...')
                eventlet.sleep(5)

    def _get_client(self, ip, port, secret, is_encrypt):
        key = self._key(ip, port, secret, is_encrypt)

        if key not in self._client_pool:
            self._client_pool[key] = self._init_client(ip, port, secret, is_encrypt)

        self._last_act_at[key] = datetime.datetime.now()
        self._is_busy[key] = True

        return self._client_pool[key]

    def _delete(self, key):
        del self._client_pool[key]
        del self._last_act_at[key]
        del self._is_busy[key]

        if self.verbose:
            print('jsonsocketclient: cleared', key)

    def _cleaner(self):
        while True:
            time.sleep(10) # cleaning interval
            now = datetime.datetime.now()

            for key in list(self._client_pool.keys()):
                if self._is_busy[key]:
                    continue

                last_act = self._last_act_at[key]

                if now - last_act > datetime.timedelta(seconds=self.client_lifetime):
                    self._delete(key)


import json
from Crypto.Cipher import AES
import random
from jwt.algorithms import HMACAlgorithm
from jwt.exceptions import DecodeError
from gaw.postoffice.exceptions import PostofficeException
import time
import base64

class ConnectionTerminated(Exception):
    def __init__(self): super(ConnectionTerminated, self).__init__('Connection Terminated')


def pad(b, block_size):
    """
    padding to blocksize according to PKCS #7
    """
    padsize = block_size - len(b) % block_size
    return b + padsize * bytes(bytearray([padsize])) # python 2 compatible

def unpad(b):
    """
    unpadding according to PKCS #7
    """
    i = bytearray([b[-1]])[0]  # python 2 compatible
    return b[:-i]

def get_random_bytes(length):
    ba = bytearray(length)
    for i in range(length):
        ba[i] = random.getrandbits(8)
    return bytes(ba)

def _sign(b, secret):
    alg = HMACAlgorithm(HMACAlgorithm.SHA256)
    key = alg.prepare_key(secret)
    signature = alg.sign(b, key)
    return signature

def _verify(b, secret, signature):
    alg = HMACAlgorithm(HMACAlgorithm.SHA256)
    key = alg.prepare_key(secret)
    if not alg.verify(b, key, signature):
        raise PostofficeException(name='DecodeError',
                                  message='signature verification failed')
    return True

def attach_signature(payload, secret):
    b_payload = payload.encode('utf-8')
    return _sign(b_payload, secret) + b_payload

def verify_signature(b, secret):
    signature, b_payload = b[:32], b[32:]
    _verify(b_payload, secret, signature)
    payload = b_payload.decode('utf-8')
    return payload

def encrypt_and_sign(payload, secret):
    '''
    vi (16), ciphertext (signature 32, payload)
    '''
    iv = get_random_bytes(16)
    cipher = AES.new(secret, AES.MODE_CBC, iv)
    signed_payload = attach_signature(payload, secret)
    ciphertext = cipher.encrypt(pad(signed_payload, 16))
    return iv + ciphertext

def decrypt_and_verify(b, secret):
    iv, ciphertext = b[:16], b[16:]
    cipher = AES.new(secret, AES.MODE_CBC, iv)
    signed_payload = unpad(cipher.decrypt(ciphertext))
    payload = verify_signature(signed_payload, secret)
    return payload

def send(socket, data, secret, is_encrypt):
    # serialize
    try:
        payload = json.dumps(data)
        if secret:
            if is_encrypt:
                to_send = encrypt_and_sign(payload, secret)
            else:
                to_send = attach_signature(payload, secret)
        else:
            to_send = payload.encode('utf-8')
    except (TypeError, ValueError) as e:
        raise PostofficeException(name='ValueError',
                                  message='You can only send JSON-serializable.')

    # send the data
    socket.send('{}\n'.format(len(to_send)).encode('ascii'))
    socket.sendall(to_send)

def recieve(socket, secret, is_encrypt):
    # read the length of the data, letter by letter until we reach EOL
    length_str = b''
    char = socket.recv(1)

    # check termination
    if not char:
        raise ConnectionTerminated()

    # got a message
    # read message length
    while char != b'\n':
        length_str += char
        char = socket.recv(1)

    bulk_length = int(length_str)

    # get the payload
    buffer = memoryview(bytearray(bulk_length))
    next_offset = 0
    while bulk_length - next_offset > 0:
        recv_size = socket.recv_into(buffer[next_offset:], bulk_length - next_offset)
        next_offset += recv_size

    # deserialize payload
    try:
        if secret:
            if is_encrypt:
                payload = decrypt_and_verify(buffer.tobytes(), secret)
            else:
                payload = verify_signature(buffer.tobytes(), secret)
        else:
            payload = buffer.tobytes().decode('utf-8')
        data = json.loads(payload)
    except (TypeError, ValueError):
        raise PostofficeException(name='ValueError',
                                  message='Data received was not in JSON or not decryptable format.')

    return data


from __future__ import print_function, absolute_import
from eventlet.green import socket
from gaw.postoffice.core import send, recieve
from gaw.postoffice.exceptions import PostofficeException
import base64

class PostofficeClient:

    def __init__(self, ip, port, secret=None, is_encrypt=False, verbose=False):
        self.ip = ip
        self.port = port
        self.secret = base64.b64decode(secret) if secret else None
        self.is_encrypt = is_encrypt
        self.verbose = verbose

        self.socket = socket.socket()
        self.socket.connect((ip, port))

    def send(self, data):
        if self.verbose:
            print('postofficeclient: sending data', data)

        send(self.socket, data, self.secret, is_encrypt=self.is_encrypt)

        if self.verbose:
            print('postofficeclient: sending done')

        response = recieve(self.socket, self.secret, is_encrypt=self.is_encrypt)

        if not response['succ']:
            raise PostofficeException(name=response['name'],
                                      message=response['message'],
                                      trace=response['trace'])

        if self.verbose:
            print('postofficeclient: receciving done')

        return response['data']


from __future__ import absolute_import
from gaw import GawClient
import time

# def time_it(fn):
#     def wrapper():
#         start_time = time.time()
#         res = fn()
#         end_time = time.time()
#
#         print('time taken:', (end_time - start_time))
#         return res
#
#     return wrapper
#
# client = MicroserviceClient('127.0.0.1', 4444, verbose=False)
# print(client.test_service.plus(1, 2))
#
# @time_it
# def run_many_times():
#     for i in range(1000):
#         client.test_service.plus(a=1, b=2)
#
# run_many_times()
client = GawClient('127.0.0.1', 5555, secret='Qx9XFxN17+zkUdcBIGZ0A1sQTkUSP4SZ', is_encrypt=True, verbose=True)
rpc = client.math_service
print(rpc.plus(a=10, b=20))
print(rpc.multiply(10, 20))


from __future__ import print_function, absolute_import
from gaw.jsonsocketserver import JsonSocketClient

class GawClient:

    def __init__(self, ip, port, secret=None, is_encrypt=False,
                 connection_lifetime = 30, verbose=False,
                 request_maker=None,
                 _state=0, _service_name=None, _method_name=None):
        self.ip = ip
        self.port = port
        self.secret = secret
        self.is_encrypt = is_encrypt
        self.connection_lifetime = connection_lifetime
        self.verbose = verbose

        if request_maker is None:
            self.request_maker = JsonSocketClient(client_lifetime=connection_lifetime, verbose=verbose)
        else:
            self.request_maker = request_maker

        self._state = _state
        self._service_name = _service_name
        self._method_name = _method_name

    def __getattr__(self, item):
        # dynamic method calling handler
        if self._state == 0:
            state = 1
            service_name = item

            return GawClient(
                ip=self.ip, port=self.port,
                secret=self.secret, is_encrypt=self.is_encrypt,
                connection_lifetime=self.connection_lifetime,
                verbose=self.verbose,
                request_maker=self.request_maker,
                _state=state,
                _service_name=service_name,
                _method_name=self._method_name
            )
        elif self._state == 1:
            method_name = item

            return self.get_procedure_caller(
                service_name=self._service_name,
                method_name=method_name,
                ip=self.ip,
                port=self.port,
                secret=self.secret,
                is_encrypt=self.is_encrypt,
                request_maker=self.request_maker,
                verbose=self.verbose
            )
        else:
            raise ValueError('state not recognized')

    def get_procedure_caller(self, service_name, method_name,
                             ip, port, secret, is_encrypt,
                             request_maker, verbose):
        assert isinstance(request_maker, JsonSocketClient), 'request maker should be a jsonsocketclient'
        path = '{}/{}'.format(service_name, method_name)

        def rpc(*args, **kwargs):
            if verbose:
                print('gawclient: procedure call path', path, 'args:', args, 'kwargs:', kwargs)

            return request_maker.request(ip=ip, port=port, path=path,
                                         secret=secret, is_encrypt=is_encrypt,
                                         payload=dict(
                                             args=args,
                                             kwargs=kwargs
                                         ))
        return rpc



class JsonSocketException(Exception):

    def __init__(self, name, message, trace):
        self.name = name
        self.message = message
        self.trace = trace

    def __str__(self):
        return '\nname: {}\nmessage: {}\ntrace: {}'.format(self.name, self.message, self.trace)


from __future__ import absolute_import
from abc import ABCMeta, abstractmethod

class Serializable:
    __metaclass__ = ABCMeta

    @abstractmethod
    def dict(self):
        raise NotImplementedError()

    def __iter__(self):
        for key, val in self.dict().items():
            yield key, val

    @classmethod
    def parse(cls, dictionary):
        raise NotImplementedError()


from __future__ import print_function, absolute_import
from gaw.entrypoint import Entrypoint
from gaw.jsonsocketserver import JsonSocketServer

def pluck(d, *args):
    assert isinstance(d, dict)
    return (d[arg] for arg in args)

class GawServer:

    def __init__(self, ip, port, secret=None, is_encrypt=False, verbose=False):
        self.ip = ip
        self.port = port
        self.secret = secret
        self.is_encrypt = is_encrypt
        self.verbose = verbose
        self.socketserver = JsonSocketServer(ip, port, secret=secret, is_encrypt=is_encrypt, verbose=verbose)

        self.path = dict()

    def add(self, service_class, *args, **kwargs):
        service_name = service_class.name

        print('gawserver: running service', service_name)

        entry_methods = Entrypoint.get_entrypoints_from_class(service_class)

        for method_name in entry_methods:
            path = '{}/{}'.format(service_name, method_name)

            if self.verbose: print('gawserver: add path {}'.format(path))

            # add path
            self.socketserver.register_route(path, self.router)

            # save misc
            self.path[path] = dict(service_name=service_name,
                                   service_class=service_class,
                                   method_name=method_name,
                                   args=args,
                                   kwargs=kwargs)

        return self

    def run(self):
        self.socketserver.start()

    def router(self, path, **payload):
        if self.verbose:
            print('gawserver: router path:', path)

        cls, method_name, args, kwargs = pluck(self.path[path], 'service_class', 'method_name', 'args', 'kwargs')

        # init a new instance of the class
        cls_instance = cls(*args, **kwargs)

        # call the designated method
        args = payload['args']
        kwargs = payload['kwargs']
        result = getattr(cls_instance, method_name)(*args, **kwargs)

        if self.verbose:
            print('gawserver: result ', result)

        return result





import types
from functools import partial
import inspect

ENTRYPOINT_ATTR = 'microservice_entrypoint'

class Entrypoint(object):

    @staticmethod
    def get_entrypoints_from_class(cls):

        def is_method(obj):
            name, val = obj
            # in Python 2 it uses MetodType, in Python 3 it uses FunctionType
            return type(getattr(cls, name)) in {types.FunctionType, types.MethodType}

        methods = list(filter(is_method, inspect.getmembers(cls)))
        entry_methods = []
        for name, method in methods:
            if getattr(method, ENTRYPOINT_ATTR, None):
                entry_methods.append(name)
        return entry_methods

    @staticmethod
    def mark_method_as_entrypoint(meth):
        old_mark = getattr(meth, ENTRYPOINT_ATTR, None)

        if old_mark:
            raise ValueError('the method already marked as entrypoint')

        setattr(meth, ENTRYPOINT_ATTR, True)

    @classmethod
    def decorator(cls, *args, **kwargs):

        def registering_decorator(fn, args, kwargs):
            cls.mark_method_as_entrypoint(fn)
            return fn

        if len(args) == 1 and isinstance(args[0], types.FunctionType):
            # usage without arguments to the decorator:
            # @foobar
            # def spam():
            #     pass
            return registering_decorator(args[0], args=(), kwargs={})
        else:
            # usage with arguments to the decorator:
            # @foobar('shrub', ...)
            # def spam():
            #     pass
            return partial(registering_decorator, args=args, kwargs=kwargs)

entrypoint = Entrypoint.decorator


from gaw.gawclient import GawClient
from gaw.gawserver import GawServer
from gaw.entrypoint import entrypoint
from gaw.jsonsocketserver.exceptions import JsonSocketException
from gaw.postoffice.exceptions import PostofficeException

__version__ = '0.5.1'


from __future__ import print_function, absolute_import
import eventlet
from gaw.postoffice.core import recieve, send, ConnectionTerminated, PostofficeException
import base64
import traceback

class PostofficeServer:

    def __init__(self, ip, port, on_message, secret=None, is_encrypt=False, verbose=False):
        print('postoffice: listening on ip:', ip, 'port:', port, 'secret:', secret, 'is_encrypt:', is_encrypt)

        self.secret = base64.b64decode(secret) if secret else None
        self.is_encrypt = is_encrypt
        self.verbose = verbose
        self.on_message = on_message

        server = eventlet.listen((ip, port))
        pool = eventlet.GreenPool()
        while True:
            try:
                socket, address = server.accept()
                pool.spawn_n(self._on_connection, socket, address)
            except (SystemExit, KeyboardInterrupt):
                break

    def _success(self, payload):
        return dict(succ=True, data=payload)

    def _fail(self, name, message, trace=None):
        return dict(succ=False, data=dict(name=name,
                                          message=message,
                                          trace=trace))

    def _on_connection(self, socket, address):
        if self.verbose:
            print('postoffice: get a connection from', address)

        # loop until connection terminated
        while True:

            try:
                request = recieve(socket, self.secret, is_encrypt=self.is_encrypt)
                if self.verbose: print('postoffice: get a message ', str(request)[:150], '...')
            except ConnectionTerminated:
                if self.verbose: print('postoffice: connection ended')
                return
            except PostofficeException as e:
                if self.verbose: print('postoffice: ', e)
                wrapped_response = self._fail(name=e.name,
                                              message=e.message,
                                              trace=e.trace)
            except Exception as e:
                if self.verbose: print('postoffice: exception: ', e)
                name = type(e).__name__
                trace = traceback.format_exc()
                wrapped_response = self._fail(name=name,
                                              message=e.__str__(),
                                              trace=trace)
            else:
                response = self.on_message(request)
                wrapped_response = self._success(response)

            send(socket, wrapped_response, self.secret, is_encrypt=self.is_encrypt)


from __future__ import print_function
from setuptools import setup
from pip.req import parse_requirements
from os.path import join, dirname, abspath
import re

def get_version_from(file):
    version = re.search(
        '^__version__\s*=\s*\'(.*)\'',
        open(file).read(),
        re.M
    ).group(1)
    return version

version = get_version_from(join(dirname(__file__), 'gaw', '__init__.py'))

setup(
    name = 'gaw',
    packages = ['gaw',
                'gaw.jsonsocketserver',
                'gaw.postoffice',
                'gaw.cli'],
    install_requires=[
        'eventlet',
        'PyJWT',
        'pycrypto'
    ],
    entry_points={
        "console_scripts": ['gaw = gaw.cli.servicerunner:servicerunner_cli']
    },
    version = version,
    description = 'A small library that helps you developing microservices over simple TCP socket with ease',
    author = 'Konpat Preechakul',
    author_email = 'the.akita.ta@gmail.com',
    url = 'https://github.com/phizaz/gaw', # use the URL to the github repo
    keywords = ['microservice'] , # arbitrary keywords
    classifiers = [],
)


from __future__ import print_function
from gaw.jsonsocketserver.datatype import RequestDataType, ResponseDataType
from gaw.postoffice import PostofficeServer
import traceback

class JsonSocketServer:

    def __init__(self, ip, port, secret=None, is_encrypt=False, verbose=False):
        self.ip = ip
        self.port = port
        self.secret = secret
        self.is_encrypt = is_encrypt
        self.verbose = verbose

        self._handle_by_route = dict()

    def register_route(self, path, handle):
        self._handle_by_route[path] = handle

    def start(self):
        PostofficeServer(self.ip, self.port, self._router,
                         secret=self.secret, is_encrypt=self.is_encrypt,
                         verbose=self.verbose)

    # private

    def _router(self, message):
        if self.verbose:
            print('jsonsocketserver: router got message:', str(message)[:100], '...')

        request = RequestDataType.parse(message)

        try:
            path = request.path
            fn = self._handle_by_route[path]
        except KeyError as e:
            print('jsonsocketserver: path not found', request.path)
            name = type(e).__name__
            trace = traceback.format_exc()
            response = ResponseDataType(
                resp_to=request.id,
                success=False,
                payload=dict(
                    name=name,
                    message='path: {} not found'.format(request.path),
                    trace=trace
                )
            )
            return response.dict()

        try:
            if isinstance(request.payload, dict):
                # params can only be a dictionary
                params = request.payload
                result = fn(path=path, **params)
            else:
                assert request.payload.strip() == ''
                result = fn(path=path)
        except Exception as e:
            # catch all error, and propagate it to the requester
            name = type(e).__name__
            message = str(e)
            trace = traceback.format_exc()

            print('jsonsocketserver: error encounter .. sending back')
            print('jsonsocketserver: \ntrace:', trace)

            response = ResponseDataType(
                resp_to=request.id,
                success=False,
                payload=dict(
                    name=name,
                    message=message,
                    trace=trace,
                )
            )
            return response.dict()

        # send response back to the requester
        response = ResponseDataType(resp_to=request.id,
                                    success=True,
                                    payload=result)
        return response.dict()


from __future__ import absolute_import
from gaw.jsonsocketserver.serializable import Serializable

class RequestDataType(Serializable):

    def __init__(self,
                 id,
                 path,
                 payload):
        self.id = id
        self.path = path
        self.payload = payload

    def dict(self):
        if hasattr(self.payload, 'dict'):
            payload = self.payload.dict()
        else:
            payload = self.payload

        return dict(id=self.id,
                    path=self.path,
                    payload=payload)

    @classmethod
    def parse(cls, d):
        assert isinstance(d, dict)
        return cls(id=d['id'],
                   path=d['path'],
                   payload=d['payload'])

class ResponseDataType(Serializable):

    def __init__(self,
                 resp_to,
                 success,
                 payload):
        self.resp_to = resp_to
        self.success = success
        self.payload = payload

    def dict(self):
        if hasattr(self.payload, 'dict'):
            payload = self.payload.dict()
        else:
            payload = self.payload

        return dict(resp_to=self.resp_to,
                    success=self.success,
                    payload=payload)

    @classmethod
    def parse(cls, d):
        assert isinstance(d, dict)
        return cls(resp_to=d['resp_to'],
                   success=d['success'],
                   payload=d['payload'])